// Fill out your copyright notice in the Description page of Project Settings.
#include "LoaderBPFunctionLibrary.h"
#include "RuntimeMeshLoader.h"
#include "IImageWrapperModule.h"
#include "IImageWrapper.h"
#include <assimp/Importer.hpp>  // C++ importer interface
#include <assimp/scene.h>       // Output data structure
#include <assimp/postprocess.h> // Post processing flags


UTexture2D* ULoaderBPFunctionLibrary::LoadTexture2D(const FString& fullFilePath)
{
	IImageWrapperModule& imgWrapperModule = FModuleManager::LoadModuleChecked<IImageWrapperModule>(FName("ImageWrapper"));
	TSharedPtr<IImageWrapper> imgWrapper = imgWrapperModule.CreateImageWrapper(EImageFormat::PNG);

	//Load From File
	TArray<uint8> rawData;
	if (!FFileHelper::LoadFileToArray(rawData, *fullFilePath))
	{
		return NULL;
		UE_LOG(LogTemp, Warning, TEXT("LoadFileToArray failed: %s"), (*fullFilePath));
	}

	//Create Texture2D
	if (imgWrapper.IsValid() && imgWrapper->SetCompressed(rawData.GetData(), rawData.Num()))
	{
		const TArray<uint8>* wrapperBGRAData = NULL;
		if (imgWrapper->GetRaw(ERGBFormat::BGRA, 8, wrapperBGRAData))
		{
			UTexture2D* tex2D = UTexture2D::CreateTransient(imgWrapper->GetWidth(), imgWrapper->GetHeight(), PF_B8G8R8A8);
			if (!tex2D)
				return NULL;

			//Upload Texture
			void* texResource = tex2D->PlatformData->Mips[0].BulkData.Lock(LOCK_READ_WRITE);
			FMemory::Memcpy(texResource, wrapperBGRAData->GetData(), wrapperBGRAData->Num());
			tex2D->PlatformData->Mips[0].BulkData.Unlock();
			tex2D->UpdateResource();

			return tex2D;
		}
	}

	return NULL;
}

// To-Do: File Path
FReturnedData ULoaderBPFunctionLibrary::LoadMesh(FString filepath, EPathType type)
{
	FReturnedData result;
	result.Success=false;
	result.MeshInfos.Empty();
	result.NumMeshes = 0;

	if (filepath.IsEmpty())
	{
		UE_LOG(LogTemp, Warning, TEXT("Runtime Mesh Loader: filepath is empty.\n"));
		return result;
	}

	std::string FileName;
	std::string ModelPath = "//Recons3DAsset//";
	std::string UnrealProjectPath;
	FileName = TCHAR_TO_UTF8(*filepath);
	
	switch (type)
	{
	case EPathType::Absolute:
		ModelPath = FileName;
		break;
	case EPathType::Relative:
		// To-Do: Check File Path
#if WITH_EDITOR
		FString AppPath = FPlatformProcess::UserSettingsDir();
		AppPath.RemoveFromEnd("Local/");
		FString fullpath = AppPath + "LocalLow/HTC Corporation/SR_Reconstruction_Output/Recons3DAsset/";
		UnrealProjectPath = TCHAR_TO_UTF8(*fullpath);
#else
		ModelPath = "Binaries/Win64/Recons3DAsset/";
		UnrealProjectPath = TCHAR_TO_UTF8(*FPaths::ProjectUserDir()) + ModelPath;
#endif
		ModelPath = UnrealProjectPath + FileName;
		break;
	}

	Assimp::Importer mImporter;


	const aiScene* mScenePtr = mImporter.ReadFile(ModelPath, aiProcess_Triangulate| aiProcess_MakeLeftHanded| aiProcess_CalcTangentSpace| aiProcess_GenSmoothNormals| aiProcess_OptimizeMeshes);

	if (mScenePtr==nullptr)
	{
		UE_LOG(LogTemp,Warning,TEXT("Runtime Mesh Loader: Read mesh file failure.\n"));
		return result;
	}

	if (mScenePtr->HasMeshes())
	{
		result.MeshInfos.SetNum(mScenePtr->mNumMeshes,false);

		for (uint32 meshID = 0; meshID < mScenePtr->mNumMeshes; ++meshID)
		{
			//deal with the relative transformation,it is broken now.
			for (uint32 m=0;m<mScenePtr->mRootNode->mNumChildren;++m)
			{
				aiMatrix4x4 tempTrans=mScenePtr->mRootNode->mChildren[m]->mTransformation;
				FMatrix tempMatrix;
				tempMatrix.M[0][0] = tempTrans.a1; tempMatrix.M[0][1] = tempTrans.b1; tempMatrix.M[0][2] = tempTrans.c1; tempMatrix.M[0][3] = tempTrans.d1;
				tempMatrix.M[1][0] = tempTrans.a2; tempMatrix.M[1][1] = tempTrans.b2; tempMatrix.M[1][2] = tempTrans.c2; tempMatrix.M[1][3] = tempTrans.d2;
				tempMatrix.M[2][0] = tempTrans.a3; tempMatrix.M[2][1] = tempTrans.b3; tempMatrix.M[2][2] = tempTrans.c3; tempMatrix.M[2][3] = tempTrans.d3;
				tempMatrix.M[3][0] = tempTrans.a4; tempMatrix.M[3][1] = tempTrans.b4; tempMatrix.M[3][2] = tempTrans.c4; tempMatrix.M[3][3] = tempTrans.d4;

				result.MeshInfos[meshID].RelativeTransform = FTransform(tempMatrix);
			}

			result.MeshInfos[meshID].MeshName = mScenePtr->mMeshes[meshID]->mName.data;

			for (uint32 vertID = 0; vertID < mScenePtr->mMeshes[meshID]->mNumVertices; ++vertID)
			{
				FVector vertex = FVector(
					mScenePtr->mMeshes[meshID]->mVertices[vertID].x, 
					mScenePtr->mMeshes[meshID]->mVertices[vertID].y, 
					mScenePtr->mMeshes[meshID]->mVertices[vertID].z);
				result.MeshInfos[meshID].Vertices.Push(vertex);

				if (mScenePtr->mMeshes[meshID]->HasNormals())
				{
					FVector normal = FVector(
						mScenePtr->mMeshes[meshID]->mNormals[vertID].x,
						mScenePtr->mMeshes[meshID]->mNormals[vertID].y,
						mScenePtr->mMeshes[meshID]->mNormals[vertID].z);
					result.MeshInfos[meshID].Normals.Push(normal);
				}
				else
				{
					result.MeshInfos[meshID].Normals.Push(FVector::ZeroVector);
				}

				if (mScenePtr->mMeshes[meshID]->HasTextureCoords(0) )
				{
					FVector2D uv = FVector2D(mScenePtr->mMeshes[meshID]->mTextureCoords[0][vertID].x, -mScenePtr->mMeshes[meshID]->mTextureCoords[0][vertID].y);
					result.MeshInfos[meshID].UV0.Add(uv);
				}

				if (mScenePtr->mMeshes[meshID]->HasTangentsAndBitangents())
				{
					FProcMeshTangent meshTangent = FProcMeshTangent(
						mScenePtr->mMeshes[meshID]->mTangents[vertID].x,
						mScenePtr->mMeshes[meshID]->mTangents[vertID].y,
						mScenePtr->mMeshes[meshID]->mTangents[vertID].z
					);
					result.MeshInfos[meshID].Tangents.Push(meshTangent);
				}

				if (mScenePtr->mMeshes[meshID]->HasVertexColors(0))
				{
					FLinearColor color = FLinearColor(
						mScenePtr->mMeshes[meshID]->mColors[0][vertID].r,
						mScenePtr->mMeshes[meshID]->mColors[0][vertID].g,
						mScenePtr->mMeshes[meshID]->mColors[0][vertID].b,
						mScenePtr->mMeshes[meshID]->mColors[0][vertID].a
					);
					result.MeshInfos[meshID].VertexColors.Push(color);
				}
			}
			
			for (uint32 faceID = 0; faceID < mScenePtr->mMeshes[meshID]->mNumFaces; ++faceID)
			{
				for (uint32 m = 0; m < mScenePtr->mMeshes[meshID]->mFaces[faceID].mNumIndices; ++m)
				{
					result.MeshInfos[meshID].Triangles.Push(mScenePtr->mMeshes[meshID]->mFaces[faceID].mIndices[m]);
				}
			}

			// Deal with Textures...
			// To-Do: Keep textures in a cache pool
			aiString path; path.Clear();
			int mtxIndex = mScenePtr->mMeshes[meshID]->mMaterialIndex;		
			if (mScenePtr->mMaterials[mtxIndex]->GetTexture(aiTextureType_DIFFUSE, 0, &path) == AI_SUCCESS)
			{				
				FString textureName = path.data;

				FString FString_UnrealProjectPath(UnrealProjectPath.c_str());
				// To-Do: Check File Path
				FString fPath = FPaths::Combine(FString_UnrealProjectPath, textureName);
				result.MeshInfos[meshID].Texture = LoadTexture2D(fPath);
				if (result.MeshInfos[meshID].Texture == NULL) UE_LOG(LogTemp, Warning, TEXT("Load Texture Fail\n"));
			}
		}

		result.NumMeshes = mScenePtr->mNumMeshes;
		result.Success = true;
	}

	return result;
}