// Fill out your copyright notice in the Description page of Project Settings.

#include "ExperienceFunctionLibrary.h"
#include "IHeadMountedDisplay.h"
#include "IXRTrackingSystem.h"
#include "Runtime/Engine/Classes/Engine/LocalPlayer.h"
#include "GameFramework/PlayerController.h"
#include "Engine.h"

float UExperienceFunctionLibrary::CaculateFOVFromPlayerCamera(const APlayerController * playerController)
{
	float fov = 90.0f;
	if (playerController != nullptr)
	{
		if (playerController->PlayerCameraManager != nullptr)
		{
			fov = playerController->PlayerCameraManager->GetFOVAngle();
		}
	}
	if (!GEngine)
	{
		return fov;
	}

	if (GEngine->XRSystem.IsValid() && GEngine->IsStereoscopic3D())
	{
		float HFOV, VFOV;
		GEngine->XRSystem->GetHMDDevice()->GetFieldOfView(HFOV, VFOV);
		if (VFOV > 0 && HFOV > 0)
		{
			fov = FMath::Max(HFOV, VFOV)*1.1f;
		}
		else
		{
			FSceneViewProjectionData projectionData;

			ULocalPlayer* player = GEngine->GetGamePlayer(GEngine->GameViewport, 0);
			player->GetProjectionData(GEngine->GameViewport->Viewport, eSSP_FULL, projectionData);

			float t = projectionData.ProjectionMatrix.M[1][1];
			fov = FMath::Atan(1.f / t)*4.f*(180 / PI)*1.1f;
		}
	}
	return fov;
}

void UExperienceFunctionLibrary::UpdatePortalCaptureTransform(USceneComponent * captureComponent, const FTransform & cameraTransform, const USceneComponent * portalScene, const USceneComponent * teleportTargetScene)
{
	if (!captureComponent)
	{
		return;
	}

	const FTransform& portalSceneTransform = portalScene->GetComponentTransform();
	const FTransform& teleportTargetTransform = teleportTargetScene->GetComponentTransform();

	FVector targetLocation = teleportTargetTransform.TransformPosition(portalSceneTransform.InverseTransformPosition(cameraTransform.GetLocation()));
	FQuat targetRotation = teleportTargetTransform.GetRotation()*(portalSceneTransform.GetRotation().Inverse()*cameraTransform.GetRotation());

	captureComponent->SetWorldLocationAndRotation(targetLocation, targetRotation);
}

void UExperienceFunctionLibrary::UpdatePortalVPMParameter(USceneCaptureComponent2D * captureComponent, UMaterialInstanceDynamic * materialInstance, const FTransform & cameraTransform)
{
	
	if (!IsValid(captureComponent) || !IsValid(materialInstance))
	{
		return;
	}

	if(!IsValid(captureComponent->TextureTarget))
	{
		return;
	}

	float captureSizeX = captureComponent->TextureTarget->GetSurfaceWidth();
	float captureSizeY = captureComponent->TextureTarget->GetSurfaceHeight();
	const FTransform& transform = cameraTransform;
	FMatrix viewMatrix = transform.ToInverseMatrixWithScale();
	FVector viewLocation = transform.GetTranslation();

	viewMatrix = viewMatrix * FMatrix(
		FPlane(0, 0, 1, 0),
		FPlane(1, 0, 0, 0),
		FPlane(0, 1, 0, 0),
		FPlane(0, 0, 0, 1));
	captureComponent->FOVAngle = CaculateFOVFromPlayerCamera();
	const float FOV = captureComponent->FOVAngle * (float)PI / 360.0f;

	float XAxisMultiplier;
	float YAxisMultiplier;

	if (captureSizeX > captureSizeY)
	{
		XAxisMultiplier = 1.0f;
		YAxisMultiplier = captureSizeX / captureSizeY;
	}
	else
	{
		XAxisMultiplier = captureSizeY / captureSizeX;
		YAxisMultiplier = 1.0f;
	}

	FMatrix ProjectionMatrix = FReversedZPerspectiveMatrix(
		FOV,
		FOV,
		XAxisMultiplier,
		YAxisMultiplier,
		10,
		1000
		);

	const FMatrix ViewProjectionMatrix = viewMatrix * ProjectionMatrix;

	FVector Xaxis = ViewProjectionMatrix.GetColumn(0);
	FVector Yaxis = ViewProjectionMatrix.GetColumn(1);
	FVector Zaxis = ViewProjectionMatrix.GetColumn(2);
	FVector Waxis = ViewProjectionMatrix.GetColumn(3);

	float XaxisW = ViewProjectionMatrix.M[3][0];
	float YaxisW = ViewProjectionMatrix.M[3][1];
	float ZaxisW = ViewProjectionMatrix.M[3][2];
	float WaxisW = ViewProjectionMatrix.M[3][3];

	materialInstance->SetVectorParameterValue("PortalVPM_Xaxis", FLinearColor(Xaxis.X, Xaxis.Y, Xaxis.Z, XaxisW));
	materialInstance->SetVectorParameterValue("PortalVPM_Yaxis", FLinearColor(Yaxis.X, Yaxis.Y, Yaxis.Z, YaxisW));
	materialInstance->SetVectorParameterValue("PortalVPM_Zaxis", FLinearColor(Zaxis.X, Zaxis.Y, Zaxis.Z, ZaxisW));
	materialInstance->SetVectorParameterValue("PortalVPM_Waxis", FLinearColor(Waxis.X, Waxis.Y, Waxis.Z, WaxisW));
}
